"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const TargetLanguage_1 = require("../TargetLanguage");
const Type_1 = require("../Type");
const RendererOptions_1 = require("../RendererOptions");
const ConvenienceRenderer_1 = require("../ConvenienceRenderer");
const Naming_1 = require("../Naming");
const Strings_1 = require("../support/Strings");
const Support_1 = require("../support/Support");
const Source_1 = require("../Source");
const TypeUtils_1 = require("../TypeUtils");
const Transformers_1 = require("../Transformers");
const collection_utils_1 = require("collection-utils");
const unicode = require("unicode-properties");
const forbiddenTypeNames = ["True", "False", "None", "Enum", "List", "Dict", "Optional", "Union", "Iterable"];
const forbiddenPropertyNames = [
    "and",
    "as",
    "assert",
    "bool",
    "break",
    "class",
    "continue",
    "datetime",
    "def",
    "del",
    "elif",
    "else",
    "except",
    "finally",
    "float",
    "for",
    "from",
    "global",
    "if",
    "import",
    "in",
    "int",
    "is",
    "lambda",
    "nonlocal",
    "not",
    "or",
    "pass",
    "print",
    "raise",
    "return",
    "str",
    "try",
    "while",
    "with",
    "yield"
];
exports.pythonOptions = {
    features: new RendererOptions_1.EnumOption("python-version", "Python version", [
        ["2.7", { version: 2, typeHints: false }],
        ["3.5", { version: 3, typeHints: false }],
        ["3.6", { version: 3, typeHints: true }]
    ], "3.6")
};
class PythonTargetLanguage extends TargetLanguage_1.TargetLanguage {
    getOptions() {
        return [exports.pythonOptions.features];
    }
    get stringTypeMapping() {
        const mapping = new Map();
        // No Python programmer apparently ever needed to parse ISO date/times.
        // It's only supported in Python 3.7, which isn't out yet.
        const dateTimeType = "string";
        mapping.set("date", dateTimeType);
        mapping.set("time", dateTimeType);
        mapping.set("date-time", dateTimeType);
        // FIXME: Implement transformers
        // mapping.set("integer-string", "integer-string");
        return mapping;
    }
    get supportsUnionsWithBothNumberTypes() {
        return true;
    }
    get supportsOptionalClassProperties() {
        return false;
    }
    needsTransformerForType(t) {
        if (t instanceof Type_1.UnionType) {
            return collection_utils_1.iterableSome(t.members, m => this.needsTransformerForType(m));
        }
        if (t instanceof Type_1.ArrayType) {
            return this.needsTransformerForType(t.items);
        }
        return t.kind !== "string" && Type_1.isPrimitiveStringTypeKind(t.kind);
    }
    makeRenderer(renderContext, untypedOptionValues) {
        return new JSONPythonRenderer(this, renderContext, RendererOptions_1.getOptionValues(exports.pythonOptions, untypedOptionValues));
    }
}
exports.PythonTargetLanguage = PythonTargetLanguage;
function isStartCharacter2(utf16Unit) {
    return Strings_1.isAscii(utf16Unit) && Strings_1.isLetterOrUnderscore(utf16Unit);
}
function isPartCharacter2(utf16Unit) {
    return Strings_1.isAscii(utf16Unit) && Strings_1.isLetterOrUnderscoreOrDigit(utf16Unit);
}
function isNormalizedStartCharacter3(utf16Unit) {
    // FIXME: add Other_ID_Start - https://docs.python.org/3/reference/lexical_analysis.html#identifiers
    if (utf16Unit === 0x5f)
        return true;
    const category = unicode.getCategory(utf16Unit);
    return ["Lu", "Ll", "Lt", "Lm", "Lo", "Nl"].indexOf(category) >= 0;
}
function isNormalizedPartCharacter3(utf16Unit) {
    // FIXME: add Other_ID_Continue - https://docs.python.org/3/reference/lexical_analysis.html#identifiers
    if (isNormalizedStartCharacter3(utf16Unit))
        return true;
    const category = unicode.getCategory(utf16Unit);
    return ["Mn", "Mc", "Nd", "Pc"].indexOf(category) >= 0;
}
function isStartCharacter3(utf16Unit) {
    const s = String.fromCharCode(utf16Unit).normalize("NFKC");
    const l = s.length;
    if (l === 0 || !isNormalizedStartCharacter3(s.charCodeAt(0)))
        return false;
    for (let i = 1; i < l; i++) {
        if (!isNormalizedPartCharacter3(s.charCodeAt(i)))
            return false;
    }
    return true;
}
function isPartCharacter3(utf16Unit) {
    const s = String.fromCharCode(utf16Unit).normalize("NFKC");
    const l = s.length;
    for (let i = 0; i < l; i++) {
        if (!isNormalizedPartCharacter3(s.charCodeAt(i)))
            return false;
    }
    return true;
}
const legalizeName2 = Strings_1.utf16LegalizeCharacters(isPartCharacter2);
const legalizeName3 = Strings_1.utf16LegalizeCharacters(isPartCharacter3);
function classNameStyle(version, original) {
    const words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, version === 2 ? legalizeName2 : legalizeName3, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allUpperWordStyle, Strings_1.allUpperWordStyle, "", version === 2 ? isStartCharacter2 : isStartCharacter3);
}
function snakeNameStyle(version, original, uppercase) {
    const wordStyle = uppercase ? Strings_1.allUpperWordStyle : Strings_1.allLowerWordStyle;
    const words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, version === 2 ? legalizeName2 : legalizeName3, wordStyle, wordStyle, wordStyle, wordStyle, "_", isStartCharacter3);
}
class PythonRenderer extends ConvenienceRenderer_1.ConvenienceRenderer {
    constructor(targetLanguage, renderContext, pyOptions) {
        super(targetLanguage, renderContext);
        this.pyOptions = pyOptions;
        this.imports = new Map();
        this.declaredTypes = new Set();
    }
    forbiddenNamesForGlobalNamespace() {
        return forbiddenTypeNames;
    }
    forbiddenForObjectProperties(_, _classNamed) {
        return { names: forbiddenPropertyNames, includeGlobalForbidden: false };
    }
    makeNamedTypeNamer() {
        return Naming_1.funPrefixNamer("type", s => classNameStyle(this.pyOptions.features.version, s));
    }
    namerForObjectProperty() {
        return Naming_1.funPrefixNamer("property", s => snakeNameStyle(this.pyOptions.features.version, s, false));
    }
    makeUnionMemberNamer() {
        return null;
    }
    makeEnumCaseNamer() {
        return Naming_1.funPrefixNamer("enum-case", s => snakeNameStyle(this.pyOptions.features.version, s, true));
    }
    get commentLineStart() {
        return "# ";
    }
    emitDescriptionBlock(lines) {
        if (lines.length === 1) {
            this.emitLine('"""', lines[0], '"""');
        }
        else {
            this.emitCommentLines(lines, "", undefined, '"""', '"""');
        }
    }
    get needsTypeDeclarationBeforeUse() {
        return true;
    }
    canBeForwardDeclared(t) {
        const kind = t.kind;
        return kind === "class" || kind === "enum";
    }
    emitBlock(line, f) {
        this.emitLine(line);
        this.indent(f);
    }
    string(s) {
        const openQuote = this.pyOptions.features.version === 2 ? 'u"' : '"';
        return [openQuote, Strings_1.stringEscape(s), '"'];
    }
    withImport(module, name) {
        if (this.pyOptions.features.typeHints || module !== "typing") {
            // FIXME: This is ugly.  We should rather not generate that import in the first
            // place, but right now we just make the type source and then throw it away.  It's
            // not a performance issue, so it's fine, I just bemoan this special case, and
            // potential others down the road.
            collection_utils_1.mapUpdateInto(this.imports, module, s => (s ? collection_utils_1.setUnionInto(s, [name]) : new Set([name])));
        }
        return name;
    }
    withTyping(name) {
        return this.withImport("typing", name);
    }
    namedType(t) {
        const name = this.nameForNamedType(t);
        if (this.declaredTypes.has(t))
            return name;
        return ["'", name, "'"];
    }
    pythonType(t) {
        const actualType = Transformers_1.followTargetType(t);
        return TypeUtils_1.matchType(actualType, _anyType => this.withTyping("Any"), _nullType => "None", _boolType => "bool", _integerType => "int", _doubletype => "float", _stringType => "str", arrayType => [this.withTyping("List"), "[", this.pythonType(arrayType.items), "]"], classType => this.namedType(classType), mapType => [this.withTyping("Dict"), "[str, ", this.pythonType(mapType.values), "]"], enumType => this.namedType(enumType), unionType => {
            const maybeNullable = TypeUtils_1.nullableFromUnion(unionType);
            if (maybeNullable !== null) {
                return [this.withTyping("Optional"), "[", this.pythonType(maybeNullable), "]"];
            }
            const memberTypes = Array.from(unionType.sortedMembers).map(m => this.pythonType(m));
            return [this.withTyping("Union"), "[", collection_utils_1.arrayIntercalate(", ", memberTypes), "]"];
        }, transformedStringType => {
            if (transformedStringType.kind === "date-time") {
                return this.withImport("datetime", "datetime");
            }
            return Support_1.panic(`Transformed type ${transformedStringType.kind} not supported`);
        });
    }
    declarationLine(t) {
        if (t instanceof Type_1.ClassType) {
            return ["class ", this.nameForNamedType(t), ":"];
        }
        if (t instanceof Type_1.EnumType) {
            return ["class ", this.nameForNamedType(t), "(", this.withImport("enum", "Enum"), "):"];
        }
        return Support_1.panic(`Can't declare type ${t.kind}`);
    }
    declareType(t, emitter) {
        this.emitDescription(this.descriptionForType(t));
        this.emitBlock(this.declarationLine(t), () => {
            emitter();
        });
        this.declaredTypes.add(t);
    }
    emitClassMembers(_t) {
        return;
    }
    typeHint(...sl) {
        if (this.pyOptions.features.typeHints) {
            return sl;
        }
        return [];
    }
    emitClass(t) {
        this.declareType(t, () => {
            if (this.pyOptions.features.typeHints) {
                if (t.getProperties().size === 0) {
                    this.emitLine("pass");
                }
                else {
                    this.forEachClassProperty(t, "none", (name, jsonName, cp) => {
                        this.emitDescription(this.descriptionForClassProperty(t, jsonName));
                        this.emitLine(name, this.typeHint(": ", this.pythonType(cp.type)));
                    });
                }
                this.ensureBlankLine();
            }
            this.emitClassMembers(t);
        });
    }
    emitEnum(t) {
        this.declareType(t, () => {
            this.forEachEnumCase(t, "none", (name, jsonName) => {
                this.emitLine([name, " = ", this.string(jsonName)]);
            });
        });
    }
    emitDeclaration(decl) {
        if (decl.kind === "forward") {
            // We don't need forward declarations yet, since we only generate types.
        }
        else if (decl.kind === "define") {
            const t = decl.type;
            if (t instanceof Type_1.ClassType) {
                this.emitClass(t);
            }
            else if (t instanceof Type_1.EnumType) {
                this.emitEnum(t);
            }
            else if (t instanceof Type_1.UnionType) {
                return;
            }
            else {
                return Support_1.panic(`Cannot declare type ${t.kind}`);
            }
        }
        else {
            return Support_1.assertNever(decl.kind);
        }
    }
    emitImports() {
        this.imports.forEach((names, module) => {
            this.emitLine("from ", module, " import ", Array.from(names).join(", "));
        });
    }
    emitDefaultLeadingComments() {
        if (this.pyOptions.features.version === 2) {
            this.emitCommentLines(["# coding: utf-8"]);
            this.ensureBlankLine();
            if (this.haveEnums) {
                this.emitCommentLines(["# To use this code in Python 2.7 you'll have to", "    pip install enum34"]);
            }
        }
    }
    emitSupportCode() {
        return;
    }
    emitClosingCode() {
        return;
    }
    emitSourceStructure(_givenOutputFilename) {
        if (this.leadingComments !== undefined) {
            this.emitCommentLines(this.leadingComments);
        }
        else {
            this.emitDefaultLeadingComments();
        }
        const declarationLines = this.gatherSource(() => {
            this.forEachDeclaration(["interposing", 2], decl => this.emitDeclaration(decl));
        });
        const closingLines = this.gatherSource(() => this.emitClosingCode());
        const supportLines = this.gatherSource(() => this.emitSupportCode());
        this.ensureBlankLine();
        this.emitImports();
        this.ensureBlankLine(2);
        this.emitGatheredSource(supportLines);
        this.ensureBlankLine(2);
        this.emitGatheredSource(declarationLines);
        this.ensureBlankLine(2);
        this.emitGatheredSource(closingLines);
    }
}
exports.PythonRenderer = PythonRenderer;
function lambda(x, ...body) {
    return Source_1.multiWord(" ", "lambda", [x, ":"], body);
}
function callFn(fn, ...args) {
    return [Source_1.parenIfNeeded(fn), "(", collection_utils_1.arrayIntercalate(", ", args), ")"];
}
const identityFn = lambda("x", "x");
class JSONPythonRenderer extends PythonRenderer {
    constructor() {
        super(...arguments);
        this._deserializerFunctions = new Set();
        this._converterNamer = Naming_1.funPrefixNamer("converter", s => snakeNameStyle(this.pyOptions.features.version, s, false));
        this._topLevelConverterNames = new Map();
    }
    emitNoneConverter() {
        this.emitMultiline(`def from_none(x):
    assert x is None
    return x`);
    }
    emitBoolConverter() {
        this.emitMultiline(`def from_bool(x):
    assert isinstance(x, bool)
    return x`);
    }
    emitIntConverter() {
        this.emitMultiline(`def from_int(x):
    assert isinstance(x, int) and not isinstance(x, bool)
    return x`);
    }
    emitFromFloatConverter() {
        this.emitMultiline(`def from_float(x):
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)`);
    }
    emitToFloatConverter() {
        this.emitMultiline(`def to_float(x):
    assert isinstance(x, float)
    return x`);
    }
    emitStrConverter() {
        this.emitBlock("def from_str(x):", () => {
            const strType = this.pyOptions.features.version === 2 ? "(str, unicode)" : "str";
            this.emitLine("assert isinstance(x, ", strType, ")");
            this.emitLine("return x");
        });
    }
    emitToEnumConverter() {
        this.emitMultiline(`def to_enum(c, x):
    assert isinstance(x, c)
    return x.value`);
    }
    emitListConverter() {
        this.emitMultiline(`def from_list(f, x):
    assert isinstance(x, list)
    return [f(y) for y in x]`);
    }
    emitToClassConverter() {
        this.emitMultiline(`def to_class(c, x):
    assert isinstance(x, c)
    return x.to_dict()`);
    }
    emitDictConverter() {
        this.emitMultiline(`def from_dict(f, x):
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }`);
    }
    emitUnionConverter() {
        this.emitMultiline(`def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False`);
    }
    emitFromDatetimeConverter() {
        this.emitMultiline(`def from_datetime(x):
    # This is not correct.  Python <3.7 doesn't support ISO date-time
    # parsing in the standard library.  This is a kludge until we have
    # that.
    return datetime.strptime(x, "%Y-%m-%dT%H:%M:%S.%fZ")`);
    }
    emitToDatetimeConverter() {
        this.emitMultiline(`def to_datetime(x):
    assert isinstance(x, datetime)
    return x.isoformat()`);
    }
    emitConverter(cf) {
        switch (cf) {
            case "none":
                return this.emitNoneConverter();
            case "bool":
                return this.emitBoolConverter();
            case "int":
                return this.emitIntConverter();
            case "from-float":
                return this.emitFromFloatConverter();
            case "to-float":
                return this.emitToFloatConverter();
            case "str":
                return this.emitStrConverter();
            case "to-enum":
                return this.emitToEnumConverter();
            case "list":
                return this.emitListConverter();
            case "to-class":
                return this.emitToClassConverter();
            case "dict":
                return this.emitDictConverter();
            case "union":
                return this.emitUnionConverter();
            case "from-datetime":
                return this.emitFromDatetimeConverter();
            case "to-datetime":
                return this.emitToDatetimeConverter();
            default:
                return Support_1.assertNever(cf);
        }
    }
    conv(cf) {
        this._deserializerFunctions.add(cf);
        const name = cf.replace("-", "_");
        if (cf.startsWith("from-") || cf.startsWith("to-"))
            return name;
        return ["from_", name];
    }
    convFn(cf) {
        return Source_1.singleWord(this.conv(cf));
    }
    deserializerFn(t) {
        return TypeUtils_1.matchType(t, _anyType => identityFn, _nullType => this.convFn("none"), _boolType => this.convFn("bool"), _integerType => this.convFn("int"), _doubleType => this.convFn("from-float"), _stringType => this.convFn("str"), arrayType => lambda("x", this.conv("list"), "(", Source_1.parenIfNeeded(this.deserializerFn(arrayType.items)), ", x)"), classType => Source_1.singleWord(this.nameForNamedType(classType)), mapType => lambda("x", this.conv("dict"), "(", Source_1.parenIfNeeded(this.deserializerFn(mapType.values)), ", x)"), enumType => Source_1.singleWord(this.nameForNamedType(enumType)), unionType => {
            const deserializers = Array.from(unionType.members).map(m => this.deserializerFn(m).source);
            return lambda("x", this.conv("union"), "([", collection_utils_1.arrayIntercalate(", ", deserializers), "], x)");
        }, transformedStringType => {
            if (transformedStringType.kind === "date-time") {
                return lambda("x", callFn(this.convFn("to-datetime"), "x"));
            }
            return Support_1.panic(`Transformed type ${transformedStringType.kind} not supported`);
        });
    }
    deserializer(value, t) {
        return callFn(this.deserializerFn(t), value);
    }
    serializerFn(t) {
        return TypeUtils_1.matchType(t, _anyType => identityFn, _nullType => this.convFn("none"), _boolType => this.convFn("bool"), _integerType => this.convFn("int"), _doubleType => this.convFn("to-float"), _stringType => this.convFn("str"), arrayType => lambda("x", this.conv("list"), "(", Source_1.parenIfNeeded(this.serializerFn(arrayType.items)), ", x)"), classType => lambda("x", callFn(this.convFn("to-class"), this.nameForNamedType(classType), "x")), mapType => lambda("x", this.conv("dict"), "(", Source_1.parenIfNeeded(this.serializerFn(mapType.values)), ", x)"), enumType => lambda("x", callFn(this.convFn("to-enum"), this.nameForNamedType(enumType), "x")), unionType => {
            const serializers = Array.from(unionType.members).map(m => this.serializerFn(m).source);
            return Source_1.multiWord("", "lambda x: ", this.conv("union"), "([", collection_utils_1.arrayIntercalate(", ", serializers), "], x)");
        }, transformedStringType => {
            if (transformedStringType.kind === "date-time") {
                return lambda("x", "x.isoformat()");
            }
            return Support_1.panic(`Transformed type ${transformedStringType.kind} not supported`);
        });
    }
    serializer(value, t) {
        return [Source_1.parenIfNeeded(this.serializerFn(t)), "(", value, ")"];
    }
    emitClassMembers(t) {
        this.emitBlock("def __init__(self, obj):", () => {
            this.emitLine("assert isinstance(obj, dict)");
            this.forEachClassProperty(t, "none", (name, jsonName, cp) => {
                const property = ["obj.get(", this.string(jsonName), ")"];
                this.emitLine("self.", name, " = ", this.deserializer(property, cp.type));
            });
        });
        this.ensureBlankLine();
        this.emitBlock("def to_dict(self):", () => {
            this.emitLine("result = {}");
            this.forEachClassProperty(t, "none", (name, jsonName, cp) => {
                const property = ["self.", name];
                this.emitLine("result[", this.string(jsonName), "] = ", this.serializer(property, cp.type));
            });
            this.emitLine("return result");
        });
    }
    emitSupportCode() {
        const map = Array.from(this._deserializerFunctions).map(f => [f, f]);
        this.forEachWithBlankLines(map, ["interposing", 2], cf => {
            this.emitConverter(cf);
        });
    }
    makeTopLevelDependencyNames(_t, topLevelName) {
        const fromDict = new Naming_1.DependencyName(this._converterNamer, ConvenienceRenderer_1.topLevelNameOrder, l => `${l(topLevelName)}_from_dict`);
        const toDict = new Naming_1.DependencyName(this._converterNamer, ConvenienceRenderer_1.topLevelNameOrder, l => `${l(topLevelName)}_to_dict`);
        this._topLevelConverterNames.set(topLevelName, { fromDict, toDict });
        return [fromDict, toDict];
    }
    emitClosingCode() {
        this.forEachTopLevel(["interposing", 2], (t, name) => {
            const { fromDict, toDict } = Support_1.defined(this._topLevelConverterNames.get(name));
            const pythonType = this.pythonType(t);
            this.emitBlock(["def ", fromDict, "(s", this.typeHint(": str"), ")", this.typeHint(" -> ", pythonType), ":"], () => {
                this.emitLine("return ", this.deserializer("s", t));
            });
            this.ensureBlankLine(2);
            this.emitBlock(["def ", toDict, "(x", this.typeHint(": ", pythonType), ")", this.typeHint(" -> str"), ":"], () => {
                this.emitLine("return ", this.serializer("x", t));
            });
        });
    }
}
exports.JSONPythonRenderer = JSONPythonRenderer;
